# adapter.py

# Copyright 2022-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import sys
import ssl
import httpx
from httpx import Response
from typing import List, Optional
from prometheus_client.parser import text_string_to_metric_families as metric_parser

import aria.ops.adapter_logging as logging
from aria.ops.object import Identifier
from aria.ops.adapter_instance import AdapterInstance
from aria.ops.definition.adapter_definition import AdapterDefinition
from aria.ops.result import CollectResult, EndpointResult, TestResult
from aria.ops.timer import Timer
import constants
from constants import ADAPTER_KIND, ADAPTER_NAME

logger = logging.getLogger(__name__)

def _get_validated_certificate_for_host(hostname: str, valid_certs: List[dict]) -> Optional[str]:
    if not valid_certs:
        logger.warning(f"No valid certificates are present in the adapter instance. Cannot validate certificate for {hostname}.")
        return None
    logger.debug(f"List of accepted certificate specs in adapter:\n{valid_certs}")
    if ":" in hostname:
        host, port_str = hostname.split(":", maxsplit=1)
        try:
            port = int(port_str)
        except ValueError as e:
            port = 443
    else:
        host = hostname
        port = 443
    try:
        host_cert = str(ssl.get_server_certificate((host, port)))
        logger.debug(f"Certificate presented by {hostname}:\n{host_cert}")
    except ssl.SSLError as e:
        logger.warning(f"Could not retrieve certificate for {hostname}: {e}")
        return None
    for cert in valid_certs:
        if cert.get('cert_pem_string','').replace("\n", "").strip() == host_cert.replace("\n", "").strip():
            #.replace("\n", "") above is required as a workaround to valid_certs not containing newlines in the cert body for some reason when executing in a real Aria Ops instance.
            logger.debug(f"Certificate presented by {hostname} was found in the following spec:\n{cert}")
            return host_cert
    logger.warning(f"Certificate presented by {hostname} does not match any of the accepted certificates in the adapter instance.")
    return None


def _get_metrics(adapter_instance: AdapterInstance) -> List[Response]:
    logger.debug(f"Preparing to collect metrics from Adapter instance: {adapter_instance.get_json()}")
    hosts = adapter_instance.get_identifier_value(constants.HOST_IDENTIFIER)
    username = adapter_instance.get_credential_value(constants.USER_CREDENTIAL)
    password = adapter_instance.get_credential_value(constants.PASSWORD_CREDENTIAL)
    #certificate = _get_validated_certificate_for_host(host, adapter_instance.certificates)
    # Check that we have values for all the identifiers/credentials we need
    none_check = []
    if hosts is None:
        none_check.append(f"No host found with identifier \"{constants.HOST_IDENTIFIER}\"")
    else:
        hostsList = [host.strip() for host in hosts.split(',') if host.strip()]
    if username is None:
        none_check.append(f"No username found with identifier \"{constants.USER_CREDENTIAL}\"")
    if password is None:
        none_check.append(f"No password found with identifier \"{constants.PASSWORD_CREDENTIAL}\"")
    for host in hostsList:
        certificate = _get_validated_certificate_for_host(host, adapter_instance.certificates)
        if certificate is None:
            none_check.append(f"Failed to validate certificate for host \"{host}\". See logs for details.")
    if none_check:
        error_msg = f"Failed to retrieve the following required properties in adapter \"{ADAPTER_NAME}\": {', '.join(none_check)}"
        logger.error(error_msg)
        raise Exception(error_msg)
    
    responses = []
    for host in hostsList:
        metrics_url = f"{constants.HOST_PROTOCOL}://{host}/{constants.HOST_URL}"
        logger.debug(f"Retrieving metrics using URL: {metrics_url} User: {username}")

        # Fetch the metrics
        try:
            #Verify set to False to allow self-signed certificates or certificates with a hostname that doesn't match the hostname used to connect to the target.
            #Additionally, certificate verification is done manually above by comparing the certificate presented by the target to the list of accepted certificates in the adapter instance.
            with httpx.Client(auth=(username, password), verify=False) as client:
                response = client.get(metrics_url, timeout=60.0)
            logger.debug(f"Request status code: {response.status_code}")
            if response.status_code == 200:
                logger.debug("Request response line count: {}".format(response.text.count('\n')))
            responses.append(response)
        except httpx.RequestError as e:
            logger.error(f"Request error occurred during metrics collection: {e}")
            raise
        except httpx.HTTPStatusError as e:
            logger.error(f"HTTP error occurred during metrics collection: {e}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error occurred during metrics collection: {e}")
            raise
    return responses


def _process_sample(sample_object, family):
    if family.type == "histogram":
        if sample_object.name.endswith("_sum") or sample_object.name.endswith("_count"):
            metric_name = sample_object.name
        else:
            return None, None
    else:
        metric_name = family.name
    labels = sample_object.labels
    if "task" in labels:
        metric_name = f"{metric_name}-{labels['task']}"
    elif "method" in labels:
        if "resource" in labels:
            metric_name = f"{metric_name}-{labels['resource']}.{labels['method']}"
        else:
            metric_name = f"{metric_name}-{labels['method']}"
    return metric_name, sample_object.value


# Function to get adapter definition
def get_adapter_definition() -> AdapterDefinition:
    """
    The adapter definition defines the object types and attribute types (metric/property) that are present
    in a collection. Setting these object types and attribute types helps VMware Aria Operations to
    validate, process, and display the data correctly.
    :return: AdapterDefinition
    """
    with Timer(logger, "Get Adapter Definition"):
        definition = AdapterDefinition(ADAPTER_KIND, ADAPTER_NAME)

        definition.define_string_parameter(
            constants.HOST_IDENTIFIER,
            label="RaaS Hostname(s)",
            description=(
                "<p><strong>Hostnames/FQDNs/IPs:</strong>Comma-separated list of hostnames, "
                "FQDNs, or IP addresses for Aria Automation Config (RaaS) Servers.</p>"
                "<p>If a server uses a port other than the default <strong>443</strong>, include the port number "
                "in the format <code>'hostname:port'</code>.<br>Example: "
                "<code>'server1.example.com, 192.168.1.2:8080, server2'</code>.</p>"
            ),
            required=True
        )

        credential = definition.define_credential_type("metric_cred", "Credential")
        credential.define_string_parameter(constants.USER_CREDENTIAL, "User Name")
        credential.define_password_parameter(constants.PASSWORD_CREDENTIAL, "Password")

        # The key 'container_memory_limit' is a special key read by the VMware Aria Operations
        # collector to determine how much memory to allocate to the docker container running
        # this adapter. It does not need to be read inside the adapter code. However, removing
        # the definition from the object model will remove the ability to change the container
        # memory limit during the adapter's configuration, and the VMware Aria Operations collector
        # will give 1024 MB of memory to the container running the adapter instance.
        definition.define_int_parameter(
            "container_memory_limit",
            label="Adapter Memory Limit (MB)",
            description="Sets the maximum amount of memory VMware Aria Operations can "
            "allocate to the container running this adapter instance.",
            required=True,
            advanced=True,
            default=1024,
        )

        # TODO: Add object types, including identifiers, metrics, and properties
        # Define object and attribute types here
        # For example:
        # definition.define_metric('celery_tasks_queued_total', label='Celery Tasks Queued', description='Total number of queued Celery tasks', unit='count')

        logger.debug(f"Returning adapter definition: {definition.to_json()}")
        return definition


# Function to test adapter connectivity
def test(adapter_instance: AdapterInstance) -> TestResult:
    with Timer(logger, "Test"):
        result = TestResult()
        try:
            # A typical test connection will generally consist of:
            # 1. Read identifier values from adapter_instance that are required to
            #    connect to the target(s)
            # 2. Connect to the target(s), and retrieve some sample data
            # 3. Disconnect cleanly from the target (ensure this happens even if an
            #    error occurs)
            # 4. If any of the above failed, return an error, otherwise pass.

            # Test retrieving metrics
            metricResponses = _get_metrics(adapter_instance)

            errors = []
            for response in metricResponses:
                # Check if the response is successful
                if response.status_code != 200:
                    errors.append(f"Connection to '{response.url}' failed with status code: {response.status_code}")
            if errors:
                # Atleast one connection failed, set the TestResult to error
                result.with_error("\n".join(errors))
            else:
                result.is_success()
        except Exception as e:
            logger.error("Unexpected connection test error")
            logger.exception(e)
            result.with_error("Unexpected connection test error: " + repr(e))
        finally:
            logger.debug(f"Returning test result: {result.get_json()}")
            return result


# Function to collect metrics data
def collect(adapter_instance: AdapterInstance) -> CollectResult:
    with Timer(logger, "Collection"):
        result = CollectResult()
        try:
            # A typical collection will generally consist of:
            # 1. Read identifier values from adapter_instance that are required to
            #    connect to the target(s)
            # 2. Connect to the target(s), and retrieve data
            # 3. Add the data into a CollectResult's objects, properties, metrics, etc
            # 4. Disconnect cleanly from the target (ensure this happens even if an
            #    error occurs)
            # 5. Return the CollectResult.

            # Connect to the metrics endpoint and retrieve the data
            metricResponses = _get_metrics(adapter_instance)
            [response.raise_for_status() for response in metricResponses] # Will raise an HTTPError if the HTTP request returned an unsuccessful status code

            # Parse the Prometheus metrics and group them into RaaS & Salt Master objects
            global_obj = None
            raas_objects = {}
            master_objects = {}
            for metricResponse in metricResponses:
                for metric_family in metric_parser(metricResponse.text):
                    for sample in metric_family.samples:
                        labels = sample.labels
                        # Check and add 'raas_instance' if it exists in labels
                        if 'raas_instance' in labels:
                            if labels['raas_instance'] in raas_objects.keys():
                                obj = raas_objects[labels['raas_instance']]
                            else:
                                logger.debug(f"Adding additional '{constants.CONFIG_OBJECT}' object with key '{labels['raas_instance']}'.")
                                obj = result.object(ADAPTER_KIND, constants.CONFIG_OBJECT, labels['raas_instance'], [Identifier(key="Description", value="RaaS_Instance", is_part_of_uniqueness=False),Identifier(key="Name", value=labels['raas_instance'], is_part_of_uniqueness=True)])
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                                raas_objects[labels['raas_instance']] = obj
                        # Check and add 'master_id' if it exists in labels
                        elif 'master_id' in labels:
                            if labels['master_id'] in master_objects.keys():
                                obj = master_objects[labels['master_id']]
                            else:
                                logger.debug(f"Adding additional '{constants.MASTER_OBJECT}' object with key '{labels['master_id']}'.")
                                obj = result.object(ADAPTER_KIND, constants.MASTER_OBJECT, labels['master_id'], [Identifier(key="Description", value="Salt_Master_Instance", is_part_of_uniqueness=False),Identifier(key="Name", value=labels['master_id'], is_part_of_uniqueness=True)])
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                                master_objects[labels['master_id']] = obj
                        else:
                            if global_obj:
                                obj = global_obj
                            else:
                                adapterKey = adapter_instance.get_key()
                                adapterName = adapterKey.name
                                logger.debug(f"Adding additional '{constants.CONFIG_OBJECT}' object with key 'Global-{adapterName}'.")
                                obj = result.object(ADAPTER_KIND, constants.CONFIG_OBJECT, "Global-" + adapterName, [Identifier(key="Description", value="RaaS_Global", is_part_of_uniqueness=False),Identifier(key="AdapterInstance", value=adapterName, is_part_of_uniqueness=True)])
                                logger.debug(f"New Object JSON: {obj.get_json()}")
                                global_obj = obj
                        sample_name, sample_value = _process_sample(sample, metric_family)
                        if sample_name and sample_value:
                            if not obj.get_metric(sample_name):
                                #logger.debug(f"Adding sample with name '{sample_name}' and value '{sample_value}' to Object '{obj.get_key().name}'.")
                                obj.with_metric(sample_name, sample_value)
        except httpx.HTTPStatusError as e:
            logger.error("HTTP error occurred during collection")
            logger.exception(e)
            result.with_error(f"HTTP error occurred: {repr(e)}")
        except httpx.RequestError as e:
            logger.error("Request error occurred during collection")
            logger.exception(e)
            result.with_error(f"Request error occurred: {repr(e)}")
        except Exception as e:
            logger.error("Unexpected collection error")
            logger.exception(e)
            result.with_error("Unexpected collection error: " + repr(e))
        finally:
            logger.debug(f"Returning collection result {result.get_json()}")
            return result


# Function to get endpoints for SSL certificate handling
def get_endpoints(adapter_instance: AdapterInstance) -> EndpointResult:
    with Timer(logger, "Get Endpoints"):
        result = EndpointResult()
        # In the case that an SSL Certificate is needed to communicate to the target,
        # add each URL that the adapter uses here. Often this will be derived from a
        # 'host' parameter in the adapter instance. In this Adapter we don't use any
        # HTTPS connections, so we won't add any. If we did, we might do something like
        # this:
        # result.with_endpoint(adapter_instance.get_identifier_value("host"))
        #
        # Multiple endpoints can be returned, like this:
        # result.with_endpoint(adapter_instance.get_identifier_value("primary_host"))
        # result.with_endpoint(adapter_instance.get_identifier_value("secondary_host"))
        #
        # This 'get_endpoints' method will be run before the 'test' method,
        # and VMware Aria Operations will use the results to extract a certificate from
        # each URL. If the certificate is not trusted by the VMware Aria Operations
        # Trust Store, the user will be prompted to either accept or reject the
        # certificate. If it is accepted, the certificate will be added to the
        # AdapterInstance object that is passed to the 'test' and 'collect' methods.
        # Any certificate that is encountered in those methods should then be validated
        # against the certificate(s) in the AdapterInstance.

        # Dynamically retrieve the endpoint URL from the adapter instance
        hosts = adapter_instance.get_identifier_value(constants.HOST_IDENTIFIER)
        hostsList = [host.strip() for host in hosts.split(',') if host.strip()]
        for host in hostsList:
            # Dynamically retrieve the endpoint URL from the adapter instance
            metrics_url = f"{constants.HOST_PROTOCOL}://{host}/{constants.HOST_URL}"
            # Add the endpoint URL to the EndpointResult
            result.with_endpoint(metrics_url)
        
        logger.debug(f"Returning endpoints: {result.get_json()}")
        return result


# Main entry point of the adapter
def main(argv: List[str]) -> None:
    logging.setup_logging("adapter.log")
    # Start a new log file by calling 'rotate'. By default, the last five calls will be
    # retained. If the logs are not manually rotated, the 'setup_logging' call should be
    # invoked with the 'max_size' parameter set to a reasonable value, e.g.,
    # 10_489_760 (10MB).
    logging.rotate()

    logger.info(f"Running adapter code with arguments: {argv}")
    if len(argv) != 3:
        # `inputfile` and `outputfile` are always automatically appended to the
        # argument list by the server
        logger.error("Arguments must be <method> <inputfile> <ouputfile>")
        exit(1)

    method = argv[0]
    try:
        if method == "test":
            test(AdapterInstance.from_input()).send_results()
        elif method == "endpoint_urls":
            get_endpoints(AdapterInstance.from_input()).send_results()
        elif method == "collect":
            collect(AdapterInstance.from_input()).send_results()
        elif method == "adapter_definition":
            result = get_adapter_definition()
            if type(result) is AdapterDefinition:
                result.send_results()
            else:
                logger.info(
                    "get_adapter_definition method did not return an AdapterDefinition"
                )
                exit(1)
        else:
            logger.error(f"Command {method} not found")
            exit(1)
    finally:
        logger.info(Timer.graph())


if __name__ == "__main__":
    main(sys.argv[1:])
